@@ -225,7 +225,7 @@ favicon is ready."
 	  (w3m-favicon-set-image
 	   (w3m-favicon-convert
 	    (base64-decode-string (symbol-value icon)) 'ico))))
-       ((or (string-match "\\`https?://" url)
+       ((or (string-match "\\`\\(https?\\|file\\)://" url)
 	    (and (string-match "\\`about://\\(?:header\\|source\\)/https?://"
 			       url)
 		 (setq url (substring url 15))))
