@@ -1,6 +1,6 @@
 ;;; w3m-util.el --- Utility macros and functions for emacs-w3m
 
-;; Copyright (C) 2001-2014, 2016, 2017 TSUCHIYA Masatoshi <tsuchiya@namazu.org>
+;; Copyright (C) 2001-2014, 2016-2018 TSUCHIYA Masatoshi <tsuchiya@namazu.org>
 
 ;; Authors: TSUCHIYA Masatoshi <tsuchiya@namazu.org>,
 ;;          Shun-ichi GOTO     <gotoh@taiyo.co.jp>,
@@ -53,6 +53,7 @@
   (defvar w3m-pop-up-frames)
   (defvar w3m-pop-up-windows)
   (defvar w3m-popup-frame-parameters)
+  (defvar w3m-previous-session-buffer)
   (defvar w3m-refresh-timer)
   (defvar w3m-select-buffer-name)
   (defvar w3m-use-refresh)
@@ -76,7 +77,8 @@
   (autoload 'w3m-detect-coding-region
     (if (featurep 'emacs) "w3m-ems" "w3m-xmas"))
   (autoload 'w3m-fb-frame-parameter "w3m-fb")
-  (autoload 'w3m-history-restore-position "w3m-hist" nil t))
+  (autoload 'w3m-history-restore-position "w3m-hist" nil t)
+  (autoload 'w3m-mode "w3m"))
 
 ;; XEmacs 21.4 wants this.
 (defalias 'w3m-make-local-hook (if (featurep 'xemacs)
@@ -446,33 +448,26 @@ An argument of nil means kill the current buffer."
 (defun w3m-generate-new-buffer (name &optional next)
   "Create and return a buffer with a name based on NAME.
 Make the new buffer the next of the current buffer if NEXT is non-nil."
-  (when (string-match "<[0-9]+>\\'" name)
-    (setq name (substring name 0 (match-beginning 0))))
-  (let* ((w3m-fb-mode nil)
-	 (buffers (w3m-list-buffers))
-	 (regexp (concat "\\`" (regexp-quote name) "\\(?:<[0-9]+>\\)?\\'"))
-	 (siblings (delq nil
-			 (mapcar
-			  (lambda (buffer)
-			    (when (string-match regexp (buffer-name buffer))
-			      buffer))
-			  buffers)))
-	 youngers cur number num)
-    (if (and next
-	     (setq youngers (cdr (memq (setq cur (current-buffer)) siblings))))
-	(progn
-	  (setq number (1+ (w3m-buffer-number cur))
-		num (+ 1 (length youngers)
-		       (w3m-buffer-number (car (reverse youngers)))))
-	  (dolist (buffer (reverse youngers))
-	    (w3m-buffer-set-number buffer (setq num (1- num))))
-	  (setq num number)
-	  (dolist (buffer youngers)
-	    (w3m-buffer-set-number buffer (setq num (1+ num))))
-	  (generate-new-buffer (format "%s<%d>" name number)))
-      (if (setq number (w3m-buffer-number (car (nreverse siblings))))
-	  (generate-new-buffer (format "%s<%d>" name (1+ number)))
-	(generate-new-buffer name)))))
+  (when w3m-use-title-buffer-name
+    (let* ((tailbufs (let ((w3m-fb-mode nil))
+		       (let ((all-w3m-buffers (w3m-list-buffers)))
+			 (if next
+			     (memq (current-buffer) all-w3m-buffers)
+			   (last all-w3m-buffers)))))
+	   (new-buffer-number (w3m-buffer-number (car tailbufs))))
+      (when (string-match "\\*w3m\\*\\(<\\([0-9]+\\)>\\)?\\'" name)
+	(setq name "*w3m*"))
+      (when (and tailbufs new-buffer-number)
+	(let ((n (1+ new-buffer-number)))
+	  (dolist (buf (cdr tailbufs))
+	    (w3m-buffer-set-number buf (setq n (1+ n))))
+	  (setq name (format "%s<%d>" name (1+ new-buffer-number)))))))
+  (let ((prev (and (eq major-mode 'w3m-mode) (current-buffer)))
+	(new (generate-new-buffer name)))
+    (with-current-buffer new
+      (w3m-mode)
+      (setq w3m-previous-session-buffer prev))
+    new))
 
 (defun w3m-buffer-name-lessp (x y)
   "Return t if first arg buffer's name is less than second."
@@ -546,7 +541,7 @@ otherwise return an alist."
 	 (nreverse alist)))))
 
 (defun w3m-device-on-window-system-p ()
-  "Return non-nil if the selected frame is on a widnow system"
+  "Return non-nil if the selected frame is on a window system"
   (w3m-static-if (featurep 'xemacs)
       (device-on-window-system-p)
     window-system))
@@ -591,8 +586,7 @@ for not deleting frames made for aims other than emacs-w3m sessions.")
 (make-variable-buffer-local 'w3m-initial-frames)
 
 (defun w3m-popup-buffer (buffer)
-  "Pop up BUFFER as a new window or a new frame
-according to `w3m-pop-up-windows' and `w3m-pop-up-frames' (which see)."
+  "Pop up BUFFER as a new window or a new frame, per the `w3m-display-mode'."
   (let ((window (get-buffer-window buffer t))
 	oframe popup-frame-p frame pop-up-frames buffers other)
     (unless (eq window (selected-window))
@@ -675,7 +669,7 @@ Note that `after-make-frame-hook' doesn't take an argument."
   (unless frame
     (setq frame (selected-frame)))
   ;; Share the opened frame in `w3m-initial-frames' over all emacs-w3m
-  ;; buffers if `w3m-use-tab' is non-nil.  Otherwise, the frame is
+  ;; buffers if using a tabbed display mode.  Otherwise, the frame is
   ;; appended into `w3m-initial-frames' only in the current buffer.
   (with-current-buffer (window-buffer (frame-first-window frame))
     (when (eq major-mode 'w3m-mode)
@@ -695,18 +689,14 @@ Note that `after-make-frame-hook' doesn't take an argument."
 (defun w3m-delete-w3m-initial-frames (frame)
   "Delete FRAME from `w3m-initial-frames', the buffer-local variable.
 It is done when the FRAME in which emacs-w3m is running is deleted.
-This function is added to `delete-frame-hook' (`delete-frame-functions'
-is used instead in Emacs 22) or merged into the `delete-frame' function
-using `defadvice'."
+This function is added to `delete-frame-functions' or merged into
+the `delete-frame' function using `defadvice'."
   (save-current-buffer
     (dolist (buffer (w3m-list-buffers t))
       (set-buffer buffer)
       (setq w3m-initial-frames (delq frame w3m-initial-frames)))))
 
-(cond ((boundp 'delete-frame-functions)
-       (add-hook 'delete-frame-functions 'w3m-delete-w3m-initial-frames))
-      (t
-       (add-hook 'delete-frame-hook 'w3m-delete-w3m-initial-frames)))
+(add-hook 'delete-frame-functions 'w3m-delete-w3m-initial-frames)
 
 (defun w3m-delete-frames-and-windows (&optional exception)
   "Delete all frames and windows related to emacs-w3m buffers.
@@ -774,7 +764,10 @@ objects will not be deleted:
 		    one-window-p)
 		   (memq frame w3m-initial-frames)
 		   (not (eq (next-frame frame) frame)))
-	      (delete-frame frame))))))))
+	      (let ((delete-frame-functions
+		     (delq 'w3m-fb-delete-frame-buffers
+			   (copy-sequence delete-frame-functions))))
+		(delete-frame frame)))))))))
 
 ;;; Navigation:
 
@@ -1112,10 +1105,11 @@ ___<TAG ...>___
        nil))))
 
 (defun w3m-string-match-url-components-1 (string)
-  "Subroutine used by `w3m-string-match-url-components'."
+  "A last resort run when `w3m-string-match-url-components' fails."
 
-  ;; ^(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?
-  ;;  12            3  4          5       6  7        8 9
+  ;; \`(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?\'
+  ;;   1-----------1 3-----------3 5------56---------6 8-----8
+  ;;    2--------2      4-------4             7-----7    9--9
 
   (let ((md (make-vector 20 nil))
 	pt)
@@ -1149,7 +1143,7 @@ ___<TAG ...>___
 	(skip-chars-forward "^#")
 	(aset md 13 (setq pt (1- (point))))
 	(aset md 15 pt))
-      (unless (eobp)
+      (unless (or (eobp) (eq (char-after (1+ (point))) ?/))
 	(aset md 16 (1- (point)))
 	(aset md 18 (point))
 	(aset md 17 (setq pt (1- (point-max))))
@@ -1158,20 +1152,33 @@ ___<TAG ...>___
   0)
 
 (defconst w3m-url-components-regexp
-  "\\`\\(\\([^:/?#]+\\):\\)?\\(//\\([^/?#]*\\)\\)?\
-\\([^?#]*\\)\\(\\?\\([^#]*\\)\\)?\\(#\\(.*\\)\\)?\\'"
+  (eval-when-compile
+    (concat
+     "\\`\\(\\([^:/?#]+\\):\\)?\\(//\\([^/?#]*\\)\\)?"
+     ;;    1-----------------1   3-----------------3
+     ;;       2----------2            4---------4
+     "\\([^?#]*\\)\\(\\?\\([^#]*\\)\\)?\\(#\\(.*\\)\\)?\\'"
+     ;; 5--------5  6----------------6   8-----------8
+     ;;                   7-------7          9----9
+     ))
   "Regexp used for parsing a URI Reference.
 It matches the potential four components and fragment identifier of a
 URI reference.  See RFC2396, Appendix B for details.")
 
-(defmacro w3m-string-match-url-components (string)
+(defsubst w3m-string-match-url-components (string)
   "Do the same thing as `(string-match w3m-url-components-regexp STRING)'.
 But this function should work even if STRING is considerably long."
-  `(let ((string ,string))
-     (condition-case nil
-	 (string-match w3m-url-components-regexp string)
-       (error ;; Stack overflow in regexp matcher
-	(w3m-string-match-url-components-1 string)))))
+  (condition-case nil
+      (prog1
+	  (string-match w3m-url-components-regexp string)
+	;; Don't recognize "#" that "/" follows as a name attribute;
+	;; e.g. <https://melpa.org/#/w3m>
+	(and (match-beginning 9)
+	     (< (match-beginning 9) (match-end 9))
+	     (eq (aref string (match-beginning 9)) ?/)
+	     (set-match-data (nbutlast (match-data) 4))))
+    (error ;; Stack overflow in regexp matcher
+     (w3m-string-match-url-components-1 string))))
 
 (defun w3m-time-newer-p (a b)
   "Return t, if A is newer than B.  Otherwise return nil.
@@ -1236,7 +1243,8 @@ Otherwise return nil."
 
 (defun w3m-url-strip-fragment (url)
   "Remove the fragment identifier from the URL."
-  (if (string-match "\\`\\([^#]*\\)#" url)
+  (if (and (string-match "\\`\\([^#]*\\)#\\(/\\)?" url)
+	   (not (match-beginning 2)))
       (match-string 1 url)
     url))
 
